import random
import string

from jinja2 import (
    FileSystemLoader,
    Environment
)


def random_string(prefix = None, length=0):
    result = prefix if prefix else ''

    letters = string.ascii_lowercase + string.digits
    return f'{result}{"".join(random.choice(letters) for i in range(length-len(result)))}'

async def render_template(template_root: str = None, template: str = None, data: dict = {}) -> str:
    template_loader = FileSystemLoader(searchpath=template_root)
    template_env = Environment(loader=template_loader, enable_async=True)
    template = template_env.get_template(template)
    return await template.render_async(data)



from .base_storage_client import BaseStorageClient
from .base_storage_model import BaseStorageModel
from .base_data_model import BaseDataModel
from .data_filter import DataFilter


__all__ = [
    'BaseStorageClient',
    'BaseStorageModel',
    'BaseDataModel',
    'DataFilter',
]



from datetime import datetime

from pydantic import BaseModel


class BaseDataModel(BaseModel):
    object_id: str = None
    created_at: datetime
    updated_at: datetime

    def dict(self, **kwargs):
        hidden_fields = set([field for field in list(self.__fields__.keys()) if field[0] == '_'])
        kwargs.setdefault('exclude', hidden_fields)
        return super().dict(**kwargs)



from unittest import IsolatedAsyncioTestCase

from eyja.interfaces.db import BaseStorageModel
from eyja.hubs import DataHub
from eyja.errors import (
    ParseModelNamespaceError,
)


class BaseModelTest(IsolatedAsyncioTestCase):
    class ModelWithWrongNamespace(BaseStorageModel):
        _namespace = 'test'

        name: str
        article: str

    class ModelWithUnknownStorage(BaseStorageModel):
        _namespace = 'postgres:test:test_base_model:test_models'

        name: str
        article: str

    async def test_use_wrong_model(self):
        await DataHub.init()

        wrong_object = self.ModelWithWrongNamespace(
            name='Test',
            article='Test article',
        )

        with self.assertRaises(ParseModelNamespaceError):
            await wrong_object.save()

    async def test_use_model_with_wrong_storage(self):
        await DataHub.init()

        wrong_object = self.ModelWithUnknownStorage(
            name='Test',
            article='Test article',
        )

        with self.assertRaises(ParseModelNamespaceError):
            await wrong_object.save()



class DataFilter:
    page_no: int
    page_size: int
    fields: dict
    order_by: str
    order_direction: str

    def __init__(self, page_no = 0, page_size = 25, fields = {}, order_by = 'created_at', order_direction = 'asc'):
        self.page_no = page_no
        self.page_size = page_size
        self.fields = fields
        self.order_by = order_by
        self.order_direction = order_direction



from unittest import IsolatedAsyncioTestCase

from eyja.hubs.plugin_hub import PluginHub


class PluginHubTest(IsolatedAsyncioTestCase):
    async def test_load_plugin(self):
        await PluginHub.init(
            plugins=[
                'tests.packages.eyja_test_pyson_db',
            ]
        )

        self.assertIn('pyson-test', PluginHub.plugins())



from unittest import TestCase

from eyja.utils import load_class


class ImportTest(TestCase):
    def test_load_class(self):
        test_class_1 = load_class('tests.packages.dynamic_class_import.TestClass')
        test_class_2 = load_class('tests.packages.dynamic_class_import.test_module.Test2Class')
        test_class_3 = load_class('tests.packages.dynamic_class_import.test_module.test_submodule.Test3Class')
        test_class_4 = load_class('tests.packages.dynamic_class_import.test_module.test_submodule.Test4Class')

        self.assertEqual(test_class_1.name, 'test')
        self.assertEqual(test_class_2.name, 'test2')
        self.assertEqual(test_class_3.name, 'test3')
        self.assertEqual(test_class_4.name, 'test4')



from .dt import now
from .encoders import EyjaJSONEncoder
from .strings import random_string, render_template
from .imports import load_class, load_model


__all__ = [
    'now',
    'EyjaJSONEncoder',
    'random_string',
    'load_class', 'load_model',
]



from typing import List, Union

from eyja.hubs.config_hub import ConfigHub
from eyja.hubs.plugin_hub import PluginHub
from eyja.hubs.data_hub import DataHub


class Eyja:
    @classmethod
    async def init(
        cls,
        config_file: str = None,
        config: Union[str, dict] = None,
        plugins: List[str] = []
    ):
        await ConfigHub.init(
            config_file=config_file,
            config=config,
        )
        await PluginHub.init(
            plugins=plugins
        )
        await DataHub.init()

    @classmethod
    async def reset(cls):
        await ConfigHub.reset()
        await PluginHub.reset()
        await DataHub.reset()

    @classmethod
    def is_initialized(cls):
        return all([
            ConfigHub.is_initialized(),
            PluginHub.is_initialized(),
            DataHub.is_initialized(),
        ])



import importlib
import inspect

from typing import List

from eyja.hubs.base_hub import BaseHub
from eyja.hubs.config_hub import ConfigHub
from eyja.interfaces.plugins import BasePlugin


class PluginHub(BaseHub):
    _plugins_ids = None
    _plugins = {}

    @classmethod
    async def init(cls, plugins: List[str]):
        cls._plugins_ids = ConfigHub.get('plugins', []) + plugins

        for plugin_id in cls._plugins_ids:
            module = importlib.import_module(plugin_id)
            for _, module_item in inspect.getmembers(module):
                if inspect.isclass(module_item):
                    if issubclass(module_item, BasePlugin) and module_item != BasePlugin:
                        cls._plugins[module_item.name] = {
                            'type': module_item.plugin_type,
                            'cls': module_item,
                        }

        await super().init()

    @classmethod
    def plugins(cls):
        return cls._plugins

    @classmethod
    def plugins_by_type(cls, plugin_type):
        return {k:v for k, v in cls._plugins.items() if v['type'] == plugin_type}

    @classmethod
    async def reset(cls):
        cls._plugins_ids = []

        await super().reset()



from typing import List, Union

from eyja.utils import now
from eyja.hubs import DataHub
from eyja.errors import ParseModelNamespaceError

from .base_data_model import BaseDataModel
from .data_filter import DataFilter


class BaseStorageModel(BaseDataModel):
    _namespace: str = None
    _indexes: List[str] = [
        'created_at',
        'updated_at',
    ]
    _expiration = 0
    _key_template = 'object_id'
    _hidden_fields = []

    def __init__(self, **data):
        if 'object_id' not in data:
            data.setdefault('object_id', None)
            data.setdefault('created_at', now())
        data.setdefault('updated_at', now())
        super().__init__(**data)

    @classmethod
    def namespace(self) -> list:
        namespaces = str(self._namespace).split(':')
        if len(namespaces) != 4:
            raise ParseModelNamespaceError(
                'A namespace must have four parts - storage type, storage connection, objectspace, and object type - separated by ":"'
            )

        return namespaces

    async def save(self) -> None:
        self.updated_at = now()
        await DataHub.save(self)

    async def delete(self) -> None:
        await DataHub.delete(self)

    @classmethod
    async def delete_all(cls, filter: dict) -> None:
        await DataHub.delete_all(cls, filter)

    @property
    def data(self):
        obj_data = self.dict()

        return obj_data

    @property
    def cleared_data(self):
        obj_data = self.data
        for field in self._hidden_fields:
            obj_data.pop(field)

        return obj_data

    @classmethod
    async def get(cls, object_id: str) -> 'BaseStorageModel':
        return await DataHub.get(cls, object_id)

    @classmethod
    async def find(cls, filter: Union[DataFilter,dict]) -> List['BaseStorageModel']:
        if isinstance(filter, dict):
            filter = DataFilter(fields=filter)

        return await DataHub.find(cls, filter)



from setuptools import find_packages, setup


setup(
    name='eyja-internal',
    zip_safe=True,
    version='0.3.15',
    description='Smart async framework',
    url='https://gitlab.com/public.eyja.dev/eyja-internal',
    maintainer='Anton Berdnikov',
    maintainer_email='agratoth@yandex.ru',
    packages=find_packages(),
    package_dir={'eyja': 'eyja'},
    install_requires=[
        'aiofiles>=0.8.0',
        'pydantic>=1.8.2',
        'pytz>=2021.3',
        'PyYAML>=6.0',
        'jinja2>=3.0.3',
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires='>=3.8',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)



import importlib
import inspect

from eyja.hubs.config_hub import ConfigHub


def load_class(class_path):
    parts = class_path.split('.')
    module_path = '.'.join(parts[:-1])
    module = importlib.import_module(module_path)
    cls = getattr(module, parts[-1])
    if inspect.isclass(cls):
        return cls
    
    return None

def load_model(class_path, default_cls):
    model_path = ConfigHub.get(class_path)
    if not model_path:
        return default_cls    
    return load_class(model_path)



from eyja.hubs.base_hub import BaseHub
from eyja.hubs.config_hub import ConfigHub
from eyja.hubs.plugin_hub import PluginHub
from eyja.interfaces.plugins import BasePlugin
from eyja.constants.types import (
    PluginTypes,
    ConfigTypes,
)
from eyja.errors import ParseModelNamespaceError


class DataHub(BaseHub):
    _storages = {}

    @classmethod
    async def init(cls):
        cls._storage_configs = ConfigHub.get('storages')
        cls._storage_plugins = PluginHub.plugins_by_type(PluginTypes.STORAGE_CLIENT)

        await cls.init_storage()
        await super().init()

    @classmethod
    async def init_storage(cls):
        if not ConfigHub.get(ConfigTypes.STORAGE_CONFIGS):
            return

        for storage_name, storage_config in ConfigHub.get(ConfigTypes.STORAGE_CONFIGS).items():
            configs = {}

            if isinstance(storage_config, list):
                for config in storage_config:
                    configs[config['name']] = config
            elif isinstance(storage_config, dict):
                configs['default'] = storage_config

            plugin: BasePlugin = cls._storage_plugins[storage_name]['cls']
            for config_name, config_value in configs.items():
                if storage_name not in cls._storages:
                    cls._storages[storage_name] = {}
                cls._storages[storage_name][config_name] = await plugin.run(**config_value)

    @classmethod
    async def reset(cls):
        cls._storages = {}
        await super().reset()

    @classmethod
    def get_client(cls, obj):
        storage_type, storage_connection, object_space, object_type = obj.namespace()

        if len(storage_type) < 1:
            storage_type = list(cls._storages.keys())[0]

        if len(storage_connection) < 1:
            storage_connection = 'default'

        if storage_type not in cls._storages:
            raise ParseModelNamespaceError(
                'Unknown storage type'
            )

        if storage_connection not in cls._storages[storage_type]:
            raise ParseModelNamespaceError(
                f'Unknown storage connection - [{storage_connection}]'
            )

        client = cls._storages[storage_type][storage_connection]

        if len(object_space) < 1:
            object_space = client._buckets[0]

        return (client, object_space, object_type)

    @classmethod
    async def save(cls, obj):
        client, object_space, object_type = cls.get_client(obj)
        await client.save(obj, object_space, object_type)

    @classmethod
    async def delete(cls, obj):
        client, object_space, object_type = cls.get_client(obj)
        await client.delete(obj, object_space, object_type)

    @classmethod
    async def delete_all(cls, obj, filter):
        client, object_space, object_type = cls.get_client(obj)
        await client.delete_all(obj, object_space, object_type, filter)

    @classmethod
    async def get(cls, obj_cls, object_id):
        client, object_space, object_type = cls.get_client(obj_cls)
        return await client.get(obj_cls, object_space, object_type, object_id)

    @classmethod
    async def find(cls, obj_cls, filter):
        client, object_space, object_type = cls.get_client(obj_cls)
        return await client.find(obj_cls, object_space, object_type, filter)
